//////////////////////////////////////////////////////////////////////////
//
// $Id: QAvatarManager.cxx,v 1.9 2004/03/26 21:20:02 scharver Exp $
//
// Author: Chris Scharver
// Email: scharver@evl.uic.edu
// Copyright (c) 2003 Electronic Visualization Laboratory,
//                    University of Illinois at Chicago
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
//////////////////////////////////////////////////////////////////////////

#include <QUANTA/QUANTAnet_datapack_c.hxx>
#include <QUANTA/QUANTAnet_tcpReflector_c.hxx>
#include <QUANTA/QUANTAnet_udp_c.hxx>

#include <QAvatarManager.h>
#include <QAvatarFactory.h>
#include <QAvatarListener.h>

#include <assert.h>
#include <string.h>

#ifdef USE_STL_NAMESPACE
#include <iostream>
using std::cerr;
using std::cout;
using std::endl;
#else /* USE_STL_NAMESPACE */
#include <iostream.h>
#endif /* USE_STL_NAMESPACE */

const char QAvatarManager::AUX_MSG = 'a';
const char QAvatarManager::BYE_MSG = 'b';
const char QAvatarManager::HELLO_MSG = 'h';
const char QAvatarManager::TRACKER_MSG = 't';

QAvatarManager::QAvatarManager(QAvatarFactory* factory)
  : m_avatarDB(NULL), m_selfAvatar(NULL),
    m_avatarFactory(factory),
    m_hailingTCP(NULL),
    m_trackerUDP(NULL),
    m_hailingBuffer(NULL), m_trackerBuffer(NULL),
    m_noNetworking(false), m_trackerDelay(0.25), //1.0 / 15.0),
    m_tempBuffer(NULL)
{
  m_hailingTCP = new QUANTAnet_tcpReflectorClient_c;
  m_trackerUDP = new QUANTAnet_udp_c;
  // Generate an avatar ID based on the current time.
  double theTime = QUANTAnet_socketbase_c::getTimeInSecs();
  char idStr[256];
  sprintf(idStr,"%lf", theTime);

  // Hash the time into a unique 4 byte number.
  QUANTAmisc_hashWord_t hashVal = QUANTAmisc_hash(idStr, strlen(idStr)+1);
  long selfAvatarID = hashVal;
  m_selfAvatar = m_avatarFactory->createAvatar(selfAvatarID);
  assert(m_selfAvatar != NULL);
  m_avatarDB = new QUANTAmisc_hashDict<QAvatar*, long>;

  m_selfAvatar->lock();
  if ((m_selfAvatar->getHelloDataSize() == 0)
      || (m_selfAvatar->getTrackerDataSize() == 0)) {
    cerr << "WARNING: QAvatarManager detected uninitialized avatar "
         << "hello or tracker data size." << endl;
  }

  // Allocate memory for packing and sending hello and trackerdata.
  m_hailingBufSize = QUANTAnet_datapack_c::sizeof_char()
    + QUANTAnet_datapack_c::sizeof_long()
    + (m_selfAvatar->getHelloDataSize() > m_selfAvatar->getAuxDataSize()
       ? m_selfAvatar->getHelloDataSize() : m_selfAvatar->getAuxDataSize());
  cout << "AVATAR hailing size: " << m_hailingBufSize << endl;
  m_hailingBuffer = new char[m_hailingBufSize];
  m_trackerBufSize = QUANTAnet_datapack_c::sizeof_char()
    + QUANTAnet_datapack_c::sizeof_long()
    + m_selfAvatar->getTrackerDataSize();
  cout << "AVATAR tracker size: " << m_trackerBufSize << endl;
  m_trackerBuffer = new char[m_trackerBufSize];
  m_tempBuffer = new char[m_selfAvatar->getHelloDataSize()];
  m_selfAvatar->unlock();
}

QAvatarManager::~QAvatarManager()
{
  // Close the TCP and UDP connections.
  m_hailingTCP->close();
  m_trackerUDP->close();

  delete m_selfAvatar;
  delete m_avatarDB;
  delete m_hailingTCP;
  delete m_trackerUDP;
  delete[] m_hailingBuffer;
  delete[] m_trackerBuffer;
  delete[] m_tempBuffer;
}

void
QAvatarManager::checkForTimeouts()
{
  // Every 60 seconds check for dead avatars.
  m_avatarDB->lock();
  int numItems;
  QAvatar** avatarList = m_avatarDB->buildListOfEntries(&numItems);
  QAvatar* theAvatar;
  if (numItems) {
    for (int i = 0; i < numItems; i++) {
      theAvatar = avatarList[i];
      // If there are avatars that have not been refreshed in
      // 2 minutes purge them.
      if ((QUANTAnet_socketbase_c::getTimeInSecs()
           - theAvatar->getLastUpdateTime()) > 120) {
        m_avatarDB->remove(theAvatar->getID());
        cout << "AVATAR " << avatarList[i]->getID() << " timed out" << endl;
        // Call user callback
        this->notifyBye(theAvatar);
        delete theAvatar;
      }
    }
  }
  m_avatarDB->unlock();

  if (numItems)
    delete[] avatarList;
}

bool
QAvatarManager::connect(const char* server, const unsigned short& hailingPort,
                        const unsigned short& trackerPort)
{
  // Open the TCP channel for avatar hailing.
  if (m_hailingTCP->connectToServer(server, hailingPort)
      == QUANTAnet_tcpReflectorClient_c::OK) {
    cout << "AVATAR hailing channel connected to "
         << server << ':'<< hailingPort << endl;
  } else {
    m_noNetworking = true;
    cerr << "AVATAR TCP CONNECT FAILED" << endl;
    return false;
  }

  // Open the UDP channel for tracker information.
  m_trackerUDP->setSendAddress(server, trackerPort);
  if (m_trackerUDP->init() == QUANTAnet_udp_c::OK) {
    cout << "AVATAR tracker channel connected to "
         << server << ':'<< trackerPort << endl;
  } else {
    m_noNetworking = true;
    cerr << "AVATAR UDP CONNECT FAILED" << endl;
    return false;
  }
  return true;
}

QAvatar*
QAvatarManager::getSelfAvatar()
{
  assert(m_selfAvatar != NULL);
  return m_selfAvatar;
}

void
QAvatarManager::attach(QAvatarListener* listener)
{
  if (listener != NULL)
    m_listeners.push_back(listener);
}

void
QAvatarManager::detach(QAvatarListener* listener)
{
  if (listener != NULL)
    m_listeners.remove(listener);
}

void
QAvatarManager::process()
{
  if (m_noNetworking)
    return;

  char* hailData = NULL;
  int dataSize;
  int udpBytesRead;

  int status  = m_hailingTCP->read(&hailData, &dataSize,
                                   QUANTAnet_tcpReflectorClient_c::NON_BLOCKING);
  if (hailData) {
    this->handleHailingData(hailData, dataSize);
    delete[] hailData;
  }

  char* buf = new char[m_trackerBufSize];

  udpBytesRead = m_trackerUDP->receive(buf, m_trackerBufSize,
                                       QUANTAnet_udp_c::NON_BLOCKING);
  if (udpBytesRead > 0)
    this->handleTrackerData(buf, udpBytesRead);
  delete[] buf;

  this->checkForTimeouts();
}

bool
QAvatarManager::sendAux()
{
  if (m_noNetworking)
    return false;

  QUANTAnet_datapack_c datapack;
  datapack.initPack(m_hailingBuffer, m_hailingBufSize);
  datapack.packChar(QAvatarManager::AUX_MSG);
  datapack.packLong(m_selfAvatar->getID());
  m_selfAvatar->lock();
  datapack.pack(m_selfAvatar->getAuxData(), m_selfAvatar->getAuxDataSize());
  m_selfAvatar->unlock();

  int bufSize = datapack.getBufferFilledSize();
  if (m_hailingTCP->write(m_hailingBuffer, &bufSize)
      == QUANTAnet_tcpReflectorClient_c::OK)
    return true;

  return false;
}

bool
QAvatarManager::sendBye()
{
  if (m_noNetworking)
    return false;

  QUANTAnet_datapack_c datapack;
  datapack.initPack(m_hailingBuffer, m_hailingBufSize);
  datapack.packChar(QAvatarManager::BYE_MSG);
  datapack.packLong(m_selfAvatar->getID());

  int bufSize = datapack.getBufferFilledSize();
  if (m_hailingTCP->write(m_hailingBuffer, &bufSize)
      == QUANTAnet_tcpReflectorClient_c::OK)
    return true;

  return false;
}

bool
QAvatarManager::sendHello()
{
  if (m_noNetworking)
    return false;

  QUANTAnet_datapack_c datapack;
  datapack.initPack(m_hailingBuffer, m_hailingBufSize);
  datapack.packChar(QAvatarManager::HELLO_MSG);
  datapack.packLong(m_selfAvatar->getID());
  m_selfAvatar->lock();
  datapack.pack(m_selfAvatar->getHelloData(),
                m_selfAvatar->getHelloDataSize());
  m_selfAvatar->unlock();

  int bufSize = datapack.getBufferFilledSize();

  if (m_hailingTCP->write(m_hailingBuffer, &bufSize)
      == QUANTAnet_tcpReflectorClient_c::OK)
    return true;

  return false;
}

bool
QAvatarManager::sendTracker()
{
  if (m_noNetworking)
    return false;

  double currentTime = QUANTAnet_socketbase_c::getTimeInSecs();
  if ((currentTime - m_lastTimeTrackerSent) < m_trackerDelay) {
    return true;
  }

  QUANTAnet_datapack_c datapack;
  datapack.initPack(m_trackerBuffer, m_trackerBufSize);
  datapack.packChar(QAvatarManager::TRACKER_MSG);
  datapack.packLong(m_selfAvatar->getID());
  m_selfAvatar->lock();
  datapack.pack(m_selfAvatar->getTrackerData(),
                m_selfAvatar->getTrackerDataSize());
  m_selfAvatar->unlock();

  int bufSize = datapack.getBufferFilledSize();

  if (m_trackerUDP->send(m_trackerBuffer, bufSize, QUANTAnet_udp_c::NON_BLOCKING)
      != -1) {
    m_lastTimeTrackerSent = currentTime;
    return true;
  } else {
    cerr << "AVATAR tracker data was not sent!" << endl;
  }

  return false;
}

void
QAvatarManager::setTrackerUpdateDelay(const double& delay)
{
  assert(delay >= 0.0);
  m_trackerDelay = delay;
}

void
QAvatarManager::handleHailingData(char* dataBuf, const size_t& dataSize)
{
  if (dataSize == 0)
    return;

  cout << "AVATAR: HailingCB" << endl;
  m_avatarDB->lock();
 
  QUANTAnet_datapack_c datapack;
  
  // Extract command and avatar id
  datapack.initUnpack(dataBuf, dataSize);
  char command;
  datapack.unpackChar(&command);
  long avatarID;
  datapack.unpackLong(&avatarID);
 
  // Look for the avatar.
  QAvatar* theQAvatar;
  m_avatarDB->find(avatarID, theQAvatar);

  if (command == QAvatarManager::HELLO_MSG) {
    if (!theQAvatar) {

      // The avatar does not yet exist in the database, so therefore
      // it's a new avatar!
      theQAvatar = m_avatarFactory->createAvatar(avatarID);
      //, itsTrackerDataSize, itsHelloDataSize, itsAuxDataSize);

      if (!theQAvatar) {
        cerr << "AVATAR: Unable to allocate new entry for avatar." << endl;
      }

      // Refresh the timer so that this avatar does not get garbage collected.
      theQAvatar->markUpdateTime();

      // Store avatar in avatar database
      m_avatarDB->enter(avatarID, theQAvatar);

      // Unpack the avatar's hello data.
      datapack.unpack(m_tempBuffer, m_selfAvatar->getHelloDataSize());
      theQAvatar->setHelloData(m_tempBuffer, m_selfAvatar->getHelloDataSize());

      cout << "AVATAR: new avatar "<<  avatarID << " arrived" << endl;

      // Notify listers that the avatar has arrived.
      this->notifyArrive(theQAvatar);

      // Set a flag so that a hello message can be sent out.
      if (command == QAvatarManager::HELLO_MSG) {
        this->sendHello();
      }

    } else {

      // Make certain that this avatar does not get garbage collected.
      theQAvatar->markUpdateTime();

      // Grab the hello data.
      datapack.unpack(m_tempBuffer, m_selfAvatar->getHelloDataSize());

      // Is it different from previous value?
      if (memcmp(m_tempBuffer, theQAvatar->getHelloData(), m_selfAvatar->getHelloDataSize())) {

        // Update the avatar's hello data.
        theQAvatar->setHelloData(m_tempBuffer, m_selfAvatar->getHelloDataSize());

        // Send notification of new hello data.
        this->notifyHello(theQAvatar);
      }
    }

  } else if (command == QAvatarManager::BYE_MSG) {

    if (!theQAvatar) {
      m_avatarDB->unlock();
      return;
    }

    // Remove the avatar from the database, notify all listeners that
    // it has left, and then finally delete it.
    m_avatarDB->remove(avatarID);
    this->notifyBye(theQAvatar);
    delete theQAvatar;

  } else if (command == QAvatarManager::AUX_MSG) {

    if (!theQAvatar) {
      m_avatarDB->unlock();
      return;
    }

    // Make certain that this avatar does not get garbage collected.
    theQAvatar->markUpdateTime();

    datapack.unpack(m_tempBuffer, m_selfAvatar->getAuxDataSize());
    theQAvatar->setAuxData(m_tempBuffer, m_selfAvatar->getAuxDataSize());

    // Call user callback.
    this->notifyAux(theQAvatar);
  }

  m_avatarDB->unlock();
}

void
QAvatarManager::handleTrackerData(char* dataBuf, const size_t& dataSize)
{
  // This is an opportunity to send out a hello message based on
  // whether someone had already sent us a hello message.
  // Use the avatarDB's lock as the mutex for requireSendAHCHello
  //      avatarDB->lock();
  //      avatarDB->unlock();
 
  // Grab the command and avatar id.
  QUANTAnet_datapack_c datapack;
 
  datapack.initUnpack(dataBuf, dataSize);
  char command;
  datapack.unpackChar(&command);
  long avatarID;
  datapack.unpackLong(&avatarID);

  // Look for the avatar.
  QAvatar* theQAvatar;
  m_avatarDB->lock();
  m_avatarDB->find(avatarID, theQAvatar);
  m_avatarDB->unlock();

  // If the avatar is not in the DB, then ignore the packer.
  if (!theQAvatar) {
    return;
  }

  // Otherwise, retrieve the avatar tracker data.
  if (command == QAvatarManager::TRACKER_MSG) {
    theQAvatar->markUpdateTime();

    datapack.unpack(m_tempBuffer, m_selfAvatar->getTrackerDataSize());
    theQAvatar->setTrackerData(m_tempBuffer, m_selfAvatar->getTrackerDataSize());

    this->notifyTracker(theQAvatar);
  }
}

void
QAvatarManager::notifyArrive(QAvatar* theQAvatar)
{
  for (ListenerList::iterator i = m_listeners.begin();
      i != m_listeners.end(); i++) {
    (*i)->avatarArrive(theQAvatar);
  }
}

void
QAvatarManager::notifyAux(QAvatar* theQAvatar)
{
  for (ListenerList::iterator i = m_listeners.begin();
      i != m_listeners.end(); i++) {
    (*i)->avatarAux(theQAvatar);
  }
}

void
QAvatarManager::notifyBye(QAvatar* theQAvatar)
{
  for (ListenerList::iterator i = m_listeners.begin();
      i != m_listeners.end(); i++) {
    (*i)->avatarBye(theQAvatar);
  }
}

void
QAvatarManager::notifyHello(QAvatar* theQAvatar)
{
  for (ListenerList::iterator i = m_listeners.begin();
      i != m_listeners.end(); i++) {
    (*i)->avatarHello(theQAvatar);
  }
}

void
QAvatarManager::notifyTracker(QAvatar* theQAvatar)
{
  for (ListenerList::iterator i = m_listeners.begin();
      i != m_listeners.end(); i++) {
    (*i)->avatarTracker(theQAvatar);
  }
}
