//////////////////////////////////////////////////////////////////////////
//
// $Id: avatarserver.cxx,v 1.8 2004/04/02 21:44:16 scharver Exp $
//
// Author: Chris Scharver
// Email: scharver@evl.uic.edu
// Copyright (c) 2003 Electronic Visualization Laboratory,
//                    University of Illinois at Chicago
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
//////////////////////////////////////////////////////////////////////////

/* This demonstrates the use of the CAVERNplus_tcpReflector_c class.
 * This server accepts N client connections.
 * It receives data from the clients and reflects it to all other connected clie
nts.
 */

#include <QUANTA/QUANTAnet_tcpReflector_c.hxx>
#include <QUANTA/QUANTAnet_udpReflector_c.hxx>
#include <QUANTA/QUANTAts_barrier_c.hxx>
#include <QUANTA/QUANTAts_rwlock_c.hxx>
#include <QUANTA/QUANTAts_thread_c.hxx>

#include <signal.h>
#ifdef USE_STL_NAMESPACE
#include <iostream>
using std::cout;
using std::endl;
#else /* USE_STL_NAMESPACE */
#include <iostream.h>
#endif /* USE_STL_NAMESPACE */

/// Catches pressing control-c to kill the server.
void stopServer(int);
/// Performs TCP processing on a separate thread.
void* tcpThreadFunc(void*);
void* udpThreadFunc(void*);

QUANTAnet_tcpReflector_c* hailingserver;
QUANTAnet_udpReflector_c* trackerserver;

/// @name Locks for multi-threaded running status.
//@{
/// Indicates whether the threads should keep running.
volatile bool keepRunning = true;
/// Synchronizes all threads when finished running.
QUANTAts_barrier_c* runningBarrier;
/// Guards read-write access on keepRunning.
QUANTAts_rwlock_c* runningLock;
//@}

int
main(int argc, char** argv)
{
  QUANTAinit();

  // The default TCP port is 7000. The UDP port is always the next
  // availabel port number. Another port may be specified from the
  // launch arguments.
  short tcpPort = 7000;
  if (argc > 1) {
    tcpPort = strtol(argv[1], NULL, 0);
  }
  short udpPort = tcpPort+1;

  // Install signal handlers for catching the ctrl-c (SIGINT).
  signal(SIGINT, stopServer);
  signal(SIGTERM, stopServer);

  // Create the thread synchronization primitives.
  runningBarrier = new QUANTAts_barrier_c(2);
  runningLock = new QUANTAts_rwlock_c;
  // To watch debugging output from the locking object, uncomment the
  // following two lines.
  //runningLock->showDebug(true);
  //runningLock->setDebugMessage("runningLock");

  hailingserver = new QUANTAnet_tcpReflector_c;
  hailingserver->setBlockingTimeout(5);
  if (hailingserver->init(tcpPort, 64) != QUANTAnet_tcpReflector_c::OK) {
    cout << "avatarserver cannot open TCP listening port" << tcpPort << endl;
    exit(1) ;
  }

  trackerserver = new QUANTAnet_udpReflector_c;
  trackerserver->setTimeOutTime(5);
  trackerserver->setIncomingPort(udpPort) ;
  if (trackerserver->init() != QUANTAnet_udpReflector_c::OK) {
    cout << "avatarserver annot open UDP listening port" << udpPort << endl;
    exit(1) ;
  }

  // Create the TCP thread. The UDP reflector internally creates its
  // own thread, so it does not need a separate thread.
  QUANTAts_thread_c tcpThread;
  tcpThread.create(tcpThreadFunc, hailingserver);
  //threadID_2.create(udpThreadFunc, trackerserver);

  cout << "avatarserver listening on ports " << tcpPort
       << " (TCP) and "  << udpPort << " (UDP)." << endl;

  // Block until the other threads reach the barrier. This makes the
  // main process just wait without any performance penalty.
  runningBarrier->wait();

  // Just to be safe, wait until the tcpThread has finished.
  tcpThread.join();

  // It should now be safe to properly clean up everything. This is
  // important for the servers since their destructors should properly
  // shut everything down. Any connected clients should receive proper
  // notification from the server.
  delete hailingserver;
  delete trackerserver;
  delete runningBarrier;
  delete runningLock;

  cout << "avatarserver exiting" << endl;
  return 0;
}

void
stopServer(int dummy)
{
  cout << "Ctrl-C caught..." << endl;
  int retval = runningLock->writeLock();
  if (retval == 0) {
    cout << "\navatarserver shutting down..." << endl;
    keepRunning = false;
    runningLock->unlock();
    signal(SIGINT, SIG_DFL);
  } else {
    cout << "\nERROR trying to obtain runningLock! "
         << strerror(retval) << endl;
  }
}

void*
tcpThreadFunc(void* objPtr)
{
  // This is the Double-Checking Pattern that prevents dead-lock and
  // properly ensure that a variable is available for reading. As long
  // as runTCP is true, this function tries to read whether it should
  // keepRunning. It calls the TCP reflector's blocking process().
  bool runTCP = true;
  while (runTCP) {
    if (runningLock->readLock() == 0) {
      if (keepRunning) {
        // It's important to go ahead and release the lock immediately
        // in the event that the process call blocks.
        runningLock->unlock();
        reinterpret_cast<QUANTAnet_tcpReflector_c*>(objPtr)->process();
      } else {
        runTCP = false;
      }
    }
  }
  // When finished, wait for the barrier. This signals to the main
  // process that the thread function is not doing any more
  // processing.
  runningBarrier->wait();
  return NULL;
}

void*
udpThreadFunc(void* objPtr)
{
#if 1
  // Do nothing since the UDP reflector manages its own thread.
#else
  bool runUDP = true;
  while (runUDP) {
    if (runningLock->readLock() == 0) {
      if (keepRunning) {
        runningLock->unlock();
        reinterpret_cast<QUANTAnet_udpReflector_c*>(objPtr)->process();
      } else {
        runUDP = false;
      }
    }
  }
#endif
  runningBarrier->wait();
  return NULL;
}
