//////////////////////////////////////////////////////////////////////////
//
// $Id: demoavatar.cxx,v 1.11 2004/04/02 21:44:15 scharver Exp $
//
// Author: Chris Scharver
// Email: scharver@evl.uic.edu
// Copyright (c) 2003 Electronic Visualization Laboratory,
//                    University of Illinois at Chicago
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
//////////////////////////////////////////////////////////////////////////

/*
Simple non-graphic demo program to test CAVERN support template for avatars.
Run the program by typing:

demoavatar site lifetime

where:

site is the IP address of a remote IRB or another running demoavatar program.

lifetime (in seconds) is the amount of time to run this program before quiting.
By setting different times you can watch as avatars appear and disapper.

*/

#include <QUANTA/QUANTAglobals.hxx>
#include <QUANTA/QUANTAinit.hxx>
#include <QUANTA/QUANTAmisc_debug.hxx>
#include <QUANTA/QUANTAnet_datapack_c.hxx>

#include <QAvatar.h>
#include <QAvatarFactory.h>
#include <QAvatarListener.h>
#include <QAvatarManager.h>

#include <signal.h>

#ifdef USE_STL_NAMESPACE
#include <iostream>
using std::cerr;
using std::cout;
using std::endl;
#else /* USE_STL_NAMESPACE */
#include <iostream.h>
#endif /* USE_STL_NAMESPACE */

static const int HELLO_SIZE=256;
static const int TRACKER_SIZE=QUANTAnet_datapack_c::sizeof_int()*2;
static const int AUX_SIZE=256;

class DemoFactory : public QAvatarFactory
{
public:
  virtual QAvatar* createAvatar(const long& newID)
  {
    // The new avatar MUST specify the data sizes fpr hello, aux, and
    // tracker data! Without these values, the data transfer will not
    // be correct. These data sizes are essential, and as constants
    // they may only be initialized when the new QAvatar instance is
    // created by the factory.
    return new QAvatar(newID, HELLO_SIZE, AUX_SIZE, TRACKER_SIZE);
  }
};

class DemoListener : public QAvatarListener
{
public:
  virtual ~DemoListener() {}
  virtual void avatarArrive(QAvatar* avatar)
  {
    printf("WELCOME!!! AVATAR %i ARRIVED: %s\n",
           avatar->getID(), avatar->getHelloData());
  }
  virtual void avatarAux(QAvatar* avatar)
  {
    printf("AVATAR %i AUX: %s\n", avatar->getID(), avatar->getAuxData());
  }
  virtual void avatarBye(QAvatar* avatar)
  {
    printf("GOODBYE!! AVATAR %i LEFT\n", avatar->getID());
  }
  virtual void avatarHello(QAvatar* avatar)
  {
    printf("AVATAR %i HELLO: %s\n",
           avatar->getID(), avatar->getHelloData());
  }
  virtual void avatarTracker(QAvatar* avatar)
  {
    QUANTAnet_datapack_c unpacker;
    int x, y;
    unpacker.initUnpack(avatar->getTrackerData(), TRACKER_SIZE);
    unpacker.unpackInt(&x);
    unpacker.unpackInt(&y);    
    printf("AVATAR %i TRACKER: %i %i\n", avatar->getID(), x, y);
  }
};

volatile bool keepRunning = true;

void interruptHandler(int);

int
main(int argc, char** argv)
{
  if (argc < 2) {
    cerr << "Usage: server" << endl;
    return 1;
  }

  short hailingPort = 7000;
  short trackerPort = 7001;

  // Assign a signal callback to catch Ctrl-C.
  signal(SIGINT, interruptHandler);

  QUANTAinit();

  DemoFactory* avatarFactory = new DemoFactory;
  QAvatarManager* avatarManager = new QAvatarManager(avatarFactory);
  QAvatar* selfAvatar = avatarManager->getSelfAvatar();

  // Create the listener object and attach it as an observer of the
  // manager. This allows the listener to receive notifications.
  DemoListener* avatarListener = new DemoListener;
  avatarManager->attach(avatarListener);
  avatarManager->setTrackerUpdateDelay(1.0);

  // Connect to the avatar server.
  if (!avatarManager->connect(argv[1], hailingPort, trackerPort)) {
    cerr << "Unable to initialize the avatar manager!" << endl;
    delete avatarManager;
    return 1;
  }

  // Pack hello data with data about your avatar and then send it.
  char* helloData = new char[HELLO_SIZE];
  strcpy(helloData, "THIS IS HELLO DATA");
  selfAvatar->setHelloData(helloData, strlen(helloData)+1);
  avatarManager->sendHello();

  char* trackerData = new char[TRACKER_SIZE];
  char* auxData = new char[AUX_SIZE];

  int loops;
  if (argc > 2) {
    loops = atoi(argv[2]);
  } else {
    loops = 1000;
  }

  QUANTAnet_datapack_c packer;
  for (int i = 0; (i < loops) && keepRunning; i++) {

    cerr << i << endl;
    // Pack tracker data with whatever data you want and then send it. Note
    // that it is your responsibility to pack all data before calling this
    // function and unpack when received. For this example, pack the current
    // loop count and the number of remaining loops.
    packer.initPack(trackerData, TRACKER_SIZE);
    packer.packInt(i);
    packer.packInt(loops-i);
    // Update the self avatar's tracker data buffer.
    selfAvatar->setTrackerData(trackerData, TRACKER_SIZE);
    // Send the tracker data over the network to other users.
    avatarManager->sendTracker();

    // Every fourth loop iteration, send aux data.
    if ((i % 4) == 0) {
      sprintf(auxData, "THIS IS AUX DATA %d", i);
      selfAvatar->setAuxData(auxData, strlen(auxData)+1);
      avatarManager->sendAux();
    }
    avatarManager->process();
    QUANTAsleep(5);
  }

  // Shut down when the loop is finished, and make sure to send a bye
  // message so that other avatars know this client has left.
  cerr << "\nShutting down..." << endl;
  avatarManager->sendBye();

  delete avatarManager;
  delete avatarFactory;
  delete avatarListener;

  delete[] helloData;
  delete[] trackerData;
  delete[] auxData;

  return 0;
}

void
interruptHandler(int arg)
{
  cout << "Ctrl-C caught..." << endl;
  keepRunning = false;
}
